import os.path
from html import escape
from types import GeneratorType

from ovld import ovld

from .textgen import Breakable, Context, Text

# CSS for hrepr
styledir = f"{os.path.dirname(__file__)}/style"
css_nbreset = open(f"{styledir}/nbreset.css", encoding="utf-8").read()
css_hrepr = open(f"{styledir}/hrepr.css", encoding="utf-8").read()


# These tags are self-closing and cannot have children.
_void_tags = {
    "area",
    "base",
    "br",
    "col",
    "command",
    "embed",
    "hr",
    "img",
    "input",
    "keygen",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr",
}


# We do not escape string children for these tags.
_raw_tags = {"raw", "script", "style"}


# These tags do not correspond to real HTML tags. They cannot have
# attributes, and their children are simply concatenated and inlined
# in the parent.
_virtual_tags = {None, "inline", "raw"}


class Tag:
    """
    Representation of an HTML tag.

    .. note::

        Create tags using an instance of the HTML class.

    A Tag should be immutable, but you can create modified versions
    of a tag by calling it with additional attributes or children,
    or indexing it with class names:

    >>> H = HTML()
    >>> tag = H.div()
    >>> tag
    <div></div>
    >>> tag(1, 2, 3)
    <div>123</div>
    >>> tag['klass']
    <div class="klass"></div>
    >>> tag(id='dividi')
    <div id="dividi"></div>
    >>> tag['klass'](id='dividi')(1, 2, 3)
    <div class="klass" id="dividi">123</div>

    Attributes:
        name (str): The tag name (div, span, a, etc.)
        attributes (dict): A dictionary of attributes. Attributes that
            start with ``__`` are reserved for internal use.
        children (tuple): Children of this node.
        resources (tuple): List of resources needed by this node or its
            children.
    """

    specialized_tags = {}

    @classmethod
    def specialize(cls, name):
        """Return a new subclass specialized for the given tag name."""
        assert cls is Tag
        if name not in cls.specialized_tags:
            cls.specialized_tags[name] = type(f"Tag::{name}", (Tag,), {})
        return cls.specialized_tags[name]

    def __init__(self, name, attributes=None, children=None, resources=None):
        self.name = name
        self.attributes = attributes or {}
        self.children = children or ()
        self.resources = () if resources is None else tuple(resources)

    def fill(self, children=None, attributes=None, resources=None):
        if not children and not attributes and not resources:
            return self
        children = (*self.children, *children) if children else self.children
        attributes = (
            {**self.attributes, **attributes} if attributes else self.attributes
        )
        if isinstance(resources, Tag):
            resources = (resources,)
        resources = (
            tuple([*self.resources, *resources])
            if resources
            else self.resources
        )
        return type(self)(
            name=self.name,
            attributes=attributes,
            children=children,
            resources=resources,
        )

    def collect_resources(self, coll=None, exclude=None):
        coll = [] if coll is None else coll
        exclude = set() if exclude is None else exclude
        for res in self.resources:
            if res not in exclude:
                res.collect_resources(coll=coll, exclude=exclude)
                coll.append(res)
        exclude.update(self.resources)
        for child in iterate_children(self.children):
            if isinstance(child, Tag):
                child.collect_resources(coll=coll, exclude=exclude)
        return coll

    def get_attribute(self, attr, dflt):
        return self.attributes.get(attr, dflt)

    def is_virtual(self):
        return self.name in _virtual_tags

    def text_parts(self):
        is_virtual = self.is_virtual()
        escape_children = not self.attributes.get(
            "__raw", self.name in _raw_tags
        )

        def convert_attribute(k, v):
            if v is True:
                return k
            elif v is False:
                return ""
            elif isinstance(v, (tuple, frozenset)):
                res = " ".join(escape(cls) for cls in v)
            else:
                res = escape(str(v))
            return f'{k}="{res}"'

        def convert_child(c):
            if isinstance(c, Tag):
                return c.text_parts()
            elif escape_children:
                return escape(str(c))
            else:
                return str(c)

        attr = " ".join(
            convert_attribute(k, v)
            for k, v in self.attributes.items()
            if not k.startswith("__")
        )
        if attr:
            # raw tag cannot have attributes, because it's not a real tag
            assert not is_virtual
            attr = " " + attr

        children = list(map(convert_child, iterate_children(self.children)))
        if is_virtual:
            # Virtual tags just inlines their contents directly.
            return Breakable(start=None, body=children, end=None)
        if self.attributes.get("__void", self.name in _void_tags):
            assert len(self.children) == 0
            return Text(f"<{self.name}{attr} />")
        else:
            return Breakable(
                start=f"<{self.name}{attr}>",
                body=children,
                end=f"</{self.name}>",
            )

    def pretty(self, **config):
        rval, _ = self.text_parts().format(
            Context(
                tabsize=4,
                max_col=80,
                offset=0,
                line_offset=0,
                overflow="allow",
            ).replace(**config)
        )
        return rval

    def __getitem__(self, items):
        if not isinstance(items, tuple):
            items = (items,)
        assert all(isinstance(item, str) for item in items)
        classes = self.attributes.get("class", ()) + items
        return self.fill(attributes={"class": classes})

    def __call__(self, *children, **attributes):
        if len(children) > 0 and isinstance(children[0], dict):
            attributes = {**children[0], **attributes}
            children = children[1:]
        return self.fill(children, attributes)

    def __eq__(self, other):
        return (
            isinstance(other, Tag)
            and self.name == other.name
            and self.attributes == other.attributes
            and self.children == other.children
            and self.resources == other.resources
        )

    def __hash__(self):
        return hash(
            (
                self.name,
                tuple(self.attributes.items()),
                self.children,
                self.resources,
            )
        )

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.pretty(max_col=None)

    def _repr_html_(self):  # pragma: no cover
        """
        Jupyter Notebook hook to print this element as HTML.
        """
        elem = H.div(
            H.style(css_nbreset),
            *self.collect_resources(),
            H.div["hrepr"](self),
        )
        return str(elem)

    def as_page(self):
        """
        Wrap this Tag as a self-contained webpage. Create a page with
        the following structure:

        .. code-block:: html

            <!DOCTYPE html>
            <html>
              <head>
                <meta http-equiv="Content-type"
                      content="text/html"
                      charset="UTF-8" />
                {self.resources}
              </head>
              <body>
                {self}
              </body>
            </html>
        """
        H = HTML()
        utf8 = H.meta(
            {"http-equiv": "Content-type"}, content="text/html", charset="UTF-8"
        )
        return H.inline(
            H.raw("<!DOCTYPE html>"),
            H.html(H.head(utf8, *self.collect_resources()), H.body(self)),
        )


@ovld
def iterate_children(self, children: (list, tuple, GeneratorType)):
    for child in children:
        yield from self(child)


@ovld
def iterate_children(self, child: object):
    yield child


class HTML:
    """
    Tag factory:

    >>> H = HTML()
    >>> H.div()
    <div></div>
    >>> H.a(href="http://cool.cool.cool")("My cool site")
    <a href="http://cool.cool.cool">My cool site</a>
    >>> H.span["cool", "beans"]("How stylish!")
    <span class="cool beans">How stylish!</span>
    """

    def __init__(self, tag_class=Tag):
        self.tag_class = tag_class

    def __getattr__(self, tag_name):
        tag_name = tag_name.replace("_", "-")
        tag_class = self.tag_class
        if hasattr(tag_class, "specialize"):
            tag_class = tag_class.specialize(tag_name)
        return tag_class(tag_name)


H = HTML(tag_class=Tag)



"""Extended JSON generator.

This generator is meant to generate structures to embed into JavaScript code,
so it may generate invalid JSON (which is why we cannot use the json package).

In particular, interactive backends may generate something special for
functions, so that JS code can call Python functionality. By default, hjson
returns null as the representation of functions.
"""

import json
from types import FunctionType, MethodType

from ovld import ovld

from .textgen import Breakable, Sequence, Text, join


@ovld
def dump(self, d: dict):
    return Breakable(
        start="{",
        body=join(
            [Sequence(self(k), ": ", self(v)) for k, v in d.items()], sep=", "
        ),
        end="}",
    )


@ovld
def dump(self, seq: (list, tuple)):
    return Breakable(start="[", body=join(map(self, seq), sep=", "), end="]",)


@ovld
def dump(self, x: (int, float, str, bool, type(None))):
    return Text(json.dumps(x))


@ovld
def dump(self, fn: (FunctionType, MethodType)):
    return self(None)


@ovld
def dump(self, d: object):
    raise TypeError(
        f"Objects of type {type(d).__name__} cannot be JSON-serialized."
    )


def dumps(obj, **fmt):
    return dump(obj).to_string(**fmt)



from itertools import count
from types import SimpleNamespace

from ovld import ovld

from . import hjson
from .h import H, Tag

_c = count()


ABSENT = object()


def _extract_as(transform, node, new_tag, **extract):
    ns = SimpleNamespace()
    extract["type"] = None
    for attr, dflt in extract.items():
        setattr(ns, attr, node.get_attribute(attr, dflt))
    attrs = {
        attr: value
        for attr, value in node.attributes.items()
        if attr not in extract
    }
    new_node = getattr(H, new_tag or node.name).fill(
        children=(),
        attributes={attr: transform(value) for attr, value in attrs.items()},
        resources=[transform(res) for res in node.resources],
    )
    if isinstance(ns.type, str):
        new_node = new_node[f"hreprt-{ns.type}"]
    return new_node, node.children, ns


def _get_layout(ns, default="h"):
    if ns.short:
        return "s"
    elif ns.horizontal:
        return "h"
    elif ns.vertical:
        return "v"
    else:
        return default


def _format_sequence(seq, layout):
    if layout == "h" or layout == "s":
        container = H.div[f"hreprl-{layout}", "hrepr-body"]
        return container(*[H.div(x) for x in seq])
    elif layout == "v":
        table = H.table["hrepr-body"]()
        for x in seq:
            if isinstance(x, type(H.pair)):
                delimiter = x.get_attribute("delimiter", "")
                k, v = x.children
                row = H.tr(H.td(k), H.td["hrepr-delim"](delimiter), H.td(v))
            else:
                row = H.tr(H.td(x, colspan=3))
            table = table(row)
        return table
    else:  # pragma: no cover
        raise ValueError(f"layout should be 'h' or 'v', not '{layout}'")


@ovld(
    initial_state={
        "requirejs_resources": [
            H.script(
                type="text/javascript",
                src="https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.6/require.min.js",
            )
        ],
        "hjson": hjson.dumps,
    }
)
def standard_html(self, node: Tag):
    """Generate standard HTML from non-standard tags generated by hrepr."""
    target, children, data = _extract_as(
        self, node, None, constructor=None, options=None, export=None, id=None
    )
    if data.constructor is not None:
        elemid = data.id or f"_hrepr_elem{next(_c)}"
        target = target(id=elemid)
        opts = self.hjson(data.options)
        code = f"new {data.constructor}(document.getElementById('{elemid}'), {opts});"

        if data.export:
            jsnode = H.javascript(
                f"let {data.export} = {code}", export=data.export,
            )
        else:
            jsnode = H.javascript(code)

        return H.inline(
            self(target(*children)),
            self(jsnode(require=data.constructor, lazy=False)),
        )

    else:
        return type(node)(
            name=node.name,
            attributes={
                k: v
                if (
                    isinstance(v, str)
                    or k == "class"
                    or v is True
                    or v is False
                    or v is None
                )
                else self.hjson(v)
                for k, v in node.attributes.items()
            },
            children=tuple(self(x) for x in node.children),
            resources=[self(res) for res in node.resources],
        )


@ovld
def standard_html(self, node: type(H.ref)):
    _, children, data = _extract_as(self, node, "div", loop=False, num=-1)
    sym = "⟳" if data.loop else "#"
    ref = H.span["hrepr-ref"](sym, data.num)
    if node.children:
        return H.div["hrepr-refbox"](ref("="), *map(self, children))
    else:
        return ref


@ovld
def standard_html(self, node: type(H.bracketed)):
    rval, children, data = _extract_as(
        self,
        node,
        "div",
        start="(",
        end=")",
        short=False,
        horizontal=False,
        vertical=False,
    )
    layout = _get_layout(data, "h")
    body = _format_sequence(children, layout)
    return self(
        rval["hrepr-bracketed"](
            H.div["hrepr-open"](data.start),
            body,
            H.div["hrepr-close"](data.end),
        )
    )


@ovld
def standard_html(self, node: type(H.instance)):
    rval, children, data = _extract_as(
        self, node, "div", short=False, horizontal=False, vertical=False,
    )
    layout = _get_layout(data, "h")
    body = _format_sequence(children, layout)
    return self(
        rval["hrepr-instance", f"hreprl-{layout}"](
            H.div["hrepr-title"](self(data.type)), body
        )
    )


@ovld
def standard_html(self, node: type(H.pair)):
    rval, children, data = _extract_as(self, node, "div", delimiter="")
    k, v = children
    return self(rval["hrepr-pair"](k, data.delimiter, v))


@ovld
def standard_html(self, node: type(H.atom)):
    rval, children, data = _extract_as(self, node, "span", value=ABSENT)
    if data.value is not ABSENT:
        rval = rval[f"hreprv-{data.value}"]
    return self(rval(*children))


@ovld
def standard_html(self, node: type(H.defn)):
    rval, children, _ = _extract_as(self, node, "span")
    assert len(children) == 2
    key, name = children
    return self(
        rval[f"hreprk-{key}"](
            H.span["hrepr-defn-key"](key), " ", H.span["hrepr-defn-name"](name),
        )
    )


def _parse_reqs(reqs):
    reqs = [reqs] if isinstance(reqs, str) else (reqs or [])
    reqargs = ", ".join([r.split("/")[-1] for r in reqs])
    return str(reqs), reqargs


@ovld
def standard_html(self, node: type(H.javascript)):
    rval, children, data = _extract_as(
        self, node, "script", require=None, export=None, src=None, lazy=False
    )

    src = data.src
    if src is not None:
        assert not children
        assert not data.require
        assert data.export is not None
        if src.endswith(".js"):
            src += "?noext"
        rval = rval(f'requirejs.config({{paths: {{{data.export}: "{src}"}}}});')
        return rval.fill(resources=self.requirejs_resources)

    else:
        reqs, reqargs = _parse_reqs(data.require)
        if data.export is not None:
            children = [
                f"define('{data.export}', {reqs}, ({reqargs}) => {{",
                *children,
                f"\nreturn {data.export};}});",
                "" if data.lazy else f"require(['{data.export}'], _ => {{}});",
            ]
        else:
            assert not data.lazy
            children = [
                f"require({reqs}, ({reqargs}) => {{",
                *children,
                "});",
            ]
        return rval(*children).fill(resources=self.requirejs_resources)


@ovld
def standard_html(self, node: object):
    return node



import types
from collections import Counter
from dataclasses import fields as dataclass_fields
from dataclasses import is_dataclass

from ovld import OvldMC, has_attribute, meta, ovld

from . import std
from .h import H, Tag, css_hrepr
from .std import standard_html

default_string_cutoff = 20
default_bytes_cutoff = 20


ABSENT = object()


def _tn(x):
    return type(x).__name__


def _xtn(x):
    tx = type(x)
    mn = tx.__module__
    if mn == "builtins":
        return tx.__qualname__
    else:
        return f"{mn}.{tx.__qualname__}"


class Config:
    def __init__(self, cfg={}, parent=None):
        self._parent = parent
        self.__dict__.update(cfg)

    def with_config(self, cfg):
        return Config(cfg, self)

    def __getattr__(self, attr):
        # Only triggers for attributes not in __dict__
        if self._parent:
            return getattr(self._parent, attr)
        return None


class HreprState:
    def __init__(self):
        self.types_seen = set()
        self.stack = Counter()
        self.registry = {}
        self.depth = -1
        self.refs = {}

    def get_ref(self, objid):
        return self.refs.setdefault(objid, len(self.refs) + 1)

    def registered(self, objid):
        return objid in self.registry

    def register(self, objid, value):
        self.registry.setdefault(objid, value)

    def reregister(self, objid, value):
        if objid in self.registry:
            self.registry[objid] = value

    def make_refmap(self):
        rval = {}
        for objid, label in self.refs.items():
            rval[id(self.registry[objid])] = label
        return rval


class Hrepr(metaclass=OvldMC):
    @classmethod
    def make_interface(cls, **kw):
        kw = {"backend": standard_html, **kw}
        return Interface(cls, **kw)

    def __init__(
        self,
        *,
        H=H,
        config=None,
        master=None,
        preprocess=None,
        postprocess=None,
    ):
        self.H = H
        self.config = config or Config()
        self.master = master or self
        self.state = master.state if master else HreprState()
        self.preprocess = preprocess
        self.postprocess = postprocess

    def with_config(self, config):
        if not config:
            return self
        else:
            cfg = self.config.with_config(config)
            return type(self)(
                H=self.H,
                config=cfg,
                master=self.master,
                preprocess=self.preprocess,
                postprocess=self.postprocess,
            )

    def ref(self, obj, loop=False):
        num = self.state.get_ref(id(obj))
        ref = self.H.ref(loop=loop, num=num)
        if not self.config.shortrefs:
            ref = ref(self.hrepr_short(obj))
        return ref

    def global_resources(self):  # pragma: no cover
        return ()

    @ovld
    def hrepr_resources(self, cls: object):
        return []

    def hrepr_resources(self, cls: has_attribute("__hrepr_resources__")):
        return cls.__hrepr_resources__(self.H)

    @ovld.dispatch
    def hrepr(ovldcall, obj):
        self = ovldcall.obj
        rval = ovldcall.resolve(obj)(obj)
        if rval is NotImplemented:
            return self.hrepr_short(obj)
        else:
            self.state.register(id(obj), rval)
            return rval

    def hrepr(self, obj: object):
        return NotImplemented

    def hrepr(self, obj: has_attribute("__hrepr__")):
        return obj.__hrepr__(self.H, self)

    @ovld
    def hrepr_short(self, obj: object):
        clsn = _tn(obj)
        rval = self.H.atom("<", _xtn(obj), ">", type=clsn)
        self.state.register(id(obj), rval)
        return rval

    @ovld
    def hrepr_short(self, obj: has_attribute("__hrepr_short__")):
        return obj.__hrepr_short__(self.H, self)

    def __call__(self, obj, **config):
        if self.preprocess is not None:
            obj = self.preprocess(obj, self)

        self.state.skip_default = False
        runner = self.with_config(config)
        ido = id(obj)
        if self.state.stack[ido]:
            return runner.ref(obj, loop=True)

        if self.state.registered(ido) and not runner.config.norefs:
            return runner.ref(obj)

        # Push object on the stack to detect circular references
        self.state.stack[ido] += 1
        self.state.depth += 1

        if (
            runner.config.max_depth is not None
            and self.state.depth >= runner.config.max_depth
        ):
            rval = runner.hrepr_short(obj)
        else:
            rval = runner.hrepr(obj)

        if self.postprocess is not None:
            rval = self.postprocess(rval, obj, self)
            self.state.reregister(id(obj), rval)

        # Check that it's the right type
        htype = self.H.tag_class
        if not isinstance(rval, htype):
            raise TypeError(
                f"Return value of hrepr({type(obj)}) must be an "
                f"instance of {htype}, not {type(rval)}."
            )

        # Pop object from the stack
        self.state.depth -= 1
        self.state.stack[ido] -= 1

        # Collect resources for this object
        cls = type(obj)
        resources = self.hrepr_resources[cls](cls)
        rval = rval.fill(resources=resources)
        self.state.reregister(id(obj), rval)

        return rval


_remap = {}
for i in range(0x20):
    c = chr(i)
    _remap[c] = f"\\x{c.encode().hex()}"
_remap.update({"\\": "\\\\", "\r": "\\r", "\n": "\\n", "\t": "\\t"})


def _encode(s):
    return "".join(_remap.get(c, c) for c in s)


class StdHrepr(Hrepr):
    def __init__(self, *, std=std, **kw):
        super().__init__(**kw)
        self.std = std

    def global_resources(self):
        return (self.H.style(css_hrepr),)

    # Lists

    def hrepr(self, xs: list):
        return self.H.bracketed(
            *[self(x) for x in xs], start="[", end="]", type=_tn(xs),
        )

    def hrepr_short(self, xs: list):
        return self.H.bracketed(
            "...", short=True, start="[", end="]", type=_tn(xs),
        )

    # Tuples

    def hrepr(self, xs: tuple):
        return self.H.bracketed(
            *[self(x) for x in xs], start="(", end=")", type=_tn(xs),
        )

    def hrepr_short(self, xs: tuple):
        return self.H.bracketed(
            "...", short=True, start="(", end=")", type=_tn(xs),
        )

    # Sets

    def hrepr(self, xs: (set, frozenset)):
        return self.H.bracketed(
            *[self(x) for x in xs], start="{", end="}", type=_tn(xs),
        )

    def hrepr_short(self, xs: (set, frozenset)):
        return self.H.bracketed(
            "...", short=True, start="{", end="}", type=_tn(xs),
        )

    # Dictionaries

    def hrepr(self, obj: dict):
        return self.H.bracketed(
            *[
                self.H.pair(self(k), self(v), delimiter=": ",)
                for k, v in obj.items()
            ],
            type="dict",
            start="{",
            end="}",
            vertical=True,
        )

    def hrepr_short(self, xs: dict):
        return self.H.bracketed(
            "...", type="dict", start="{", end="}", short=True,
        )

    # Dataclasses

    def hrepr(self, obj: meta(is_dataclass)):
        return self.H.instance(
            *[
                self.H.pair(
                    self.H.atom(field.name, type="symbol"),
                    self(getattr(obj, field.name)),
                    delimiter="=",
                )
                for field in dataclass_fields(obj)
            ],
            type=_tn(obj),
            vertical=True,
        )

    def hrepr_short(self, obj: meta(is_dataclass)):
        return self.H.instance("...", type=_tn(obj), short=True,)

    # Exceptions

    def hrepr(self, obj: Exception):
        return self.H.instance["hrepr-error"](
            H.atom(str(obj.args[0])),
            *map(self, obj.args[1:]),
            type=_tn(obj),
            horizontal=True,
        )

    # Functions and methods

    def hrepr_short(self, obj: types.FunctionType):
        # types.LambdaType is types.FunctionType
        return self.H.defn("function", obj.__name__)

    def hrepr_short(self, obj: types.CoroutineType):
        return self.H.defn("coroutine", obj.__name__)

    def hrepr_short(self, obj: types.GeneratorType):
        return self.H.defn("generator", obj.__name__)

    def hrepr_short(self, obj: types.AsyncGeneratorType):
        return self.H.defn("async_generator", obj.__name__)

    def hrepr_short(self, obj: (types.MethodType, type([].__str__))):
        # Second one is types.MethodWrapperType but it's not exposed
        # in the types module in 3.6
        slf = obj.__self__
        slf = getattr(slf, "__name__", f"<{type(slf).__name__}>")
        return self.H.defn("method", f"{slf}.{obj.__name__}")

    def hrepr_short(self, obj: (type(object.__str__), type(dict.update))):
        # These are types.WrapperDescriptorType and types.MethodDescriptorType
        # but they are not exposed in the types module in 3.6
        objc = obj.__objclass__.__name__
        return self.H.defn("descriptor", f"{objc}.{obj.__name__}")

    def hrepr_short(self, obj: types.BuiltinMethodType):
        # types.BuiltinFunctionType is types.BuiltinMethodType
        slf = obj.__self__
        slf = getattr(slf, "__name__", f"<{type(slf).__name__}>")
        if slf == "builtins":
            # Let's not be redundant
            return self.H.defn("builtin", obj.__name__)
        else:
            return self.H.defn("builtin", f"{slf}.{obj.__name__}")

    def hrepr_short(self, obj: type):
        key = "metaclass" if issubclass(obj, type) else "class"
        return self.H.defn(key, obj.__name__)

    def hrepr_short(self, obj: types.ModuleType):
        return self.H.defn("module", obj.__name__)

    # Strings

    def hrepr(self, x: str):
        return self.H.atom(_encode(x), type="str")

    def hrepr_short(self, x: str):
        cutoff = self.config.string_cutoff or default_string_cutoff
        if len(x) > cutoff:
            x = x[: cutoff - 3] + "..."
        return self.H.atom(_encode(x), type="str")

    # Bytes

    def hrepr(self, x: bytes):
        cutoff = self.config.bytes_cutoff or default_bytes_cutoff
        if len(x) <= cutoff:
            return NotImplemented
        else:
            return self.H.atom(x.hex(), type="bytes")

    def hrepr_short(self, x: bytes):
        cutoff = self.config.bytes_cutoff or default_bytes_cutoff
        hx = x.hex()
        if len(hx) > cutoff:
            hx = hx[: cutoff - 3] + "..."
        return self.H.atom(hx, type="bytes")

    # Numbers

    def hrepr_short(self, x: (int, float)):
        return self.H.atom(str(x), type=_tn(x))

    # Booleans

    def hrepr_short(self, x: bool):
        return self.H.atom(str(x), value=x)

    # None

    def hrepr_short(self, x: type(None)):
        return self.H.atom(str(x), value=x)

    # Tags

    def hrepr(self, x: Tag):
        return x


def inject_reference_numbers(hcall, node, refmap):
    if isinstance(node, Tag):
        node.children = tuple(
            inject_reference_numbers(hcall, child, refmap)
            for child in node.children
        )
        refnum = refmap.get(id(node), None)
        if refnum is not None:
            return hcall.H.ref(node, num=refnum)
        else:
            return node
    else:
        return node


def _mix(hclass, mixins):
    if mixins:
        if isinstance(mixins, type):
            mixins = [mixins]
        hclass = hclass.create_subclass(*mixins)
    return hclass


class Interface:
    def __init__(
        self,
        hclass,
        *,
        backend,
        mixins=None,
        preprocess=None,
        postprocess=None,
        inject_references=True,
        fill_resources=True,
        **config_defaults,
    ):
        self.hclass = _mix(hclass, mixins)
        self.backend = backend
        self.preprocess = preprocess
        self.postprocess = postprocess
        self.inject_references = inject_references
        self.fill_resources = fill_resources
        self.config_defaults = config_defaults

    def copy(self):
        return type(self)(
            hclass=self.hclass,
            backend=self.backend,
            preprocess=self.preprocess,
            postprocess=self.postprocess,
            inject_references=self.inject_references,
            fill_resources=self.fill_resources,
            **self.config_defaults,
        )

    def variant(self, **options):
        return self.copy().configure(**options)

    def configure(
        self,
        *,
        hclass=None,
        mixins=None,
        backend=None,
        preprocess=ABSENT,
        postprocess=ABSENT,
        inject_references=ABSENT,
        fill_resources=ABSENT,
        **config_defaults,
    ):
        if hclass is not None:
            self.hclass = hclass
        self.hclass = _mix(self.hclass, mixins)
        if backend is not None:
            self.backend = backend
        if preprocess is not ABSENT:
            self.preprocess = preprocess
        if postprocess is not ABSENT:
            self.postprocess = postprocess
        if inject_references is not ABSENT:
            self.inject_references = inject_references
        if fill_resources is not ABSENT:
            self.fill_resources = fill_resources
        self.config_defaults.update(config_defaults)
        return self

    def page(self, *objs, file=None, **config):
        result = self(*objs, **config).as_page()
        if file is None:
            return result
        elif isinstance(file, str):  # pragma: no cover
            with open(file, "w") as f:
                f.write(str(result) + "\n")
        else:  # pragma: no cover
            file.write(str(result) + "\n")

    def __call__(self, *objs, **config):
        if config:
            return self.variant(**config)(*objs)
        else:
            hcall = self.hclass(
                H=H,
                config=Config(self.config_defaults),
                preprocess=self.preprocess,
                postprocess=self.postprocess,
            )
            if len(objs) == 1:
                rval = hcall(objs[0])
            else:
                rval = H.inline(*map(hcall, objs))
            if self.inject_references:
                rval = inject_reference_numbers(
                    hcall, rval, hcall.state.make_refmap()
                )
            if self.fill_resources:
                rval = rval.fill(resources=hcall.global_resources())
            rval = self.backend(rval)
            return rval



import math
from dataclasses import dataclass
from dataclasses import replace as dc_replace


@dataclass
class Context:
    tabsize: int
    max_col: int
    offset: int
    line_offset: int
    max_indent: int = None
    overflow: str = "allow"

    replace = dc_replace

    def __post_init__(self):
        if self.max_col is None:
            self.max_col = math.inf

    def next_line_offset(self):
        max_indent = (
            max(self.max_col - 15, 0)
            if (self.max_indent is None and self.overflow != "allow")
            else (self.max_indent or 1e32)
        )
        return min(self.line_offset + self.tabsize, max_indent)

    def indent(self):
        return " " * self.line_offset

    def format(self, x, **modifs):
        result, offset = x.format(self.replace(**modifs))
        if self.overflow != "allow":
            linepfx = r"\ " if self.overflow == "backslash" else ""
            while True:
                diff = self.max_col - offset
                if diff >= 0:
                    break
                overflow = result[diff:]
                if not overflow.strip():
                    result = result[:diff]
                    break
                else:
                    result = (
                        f"{result[:diff]}\n{self.indent()}{linepfx}{overflow}"
                    )
                offset = self.line_offset + len(overflow) + len(linepfx)

        return result, offset


def ensure_text(x):
    if isinstance(x, str):
        return Text(x)
    else:
        return x


def join(seq, sep):
    seq = list(seq)
    if not seq:
        return []
    else:
        *first, last = seq
        return [Sequence(x, sep) for x in first] + [last]


class TextFormatter:
    def to_string(self, **config):
        value, offset = self.format(
            Context(
                tabsize=4,
                max_col=None,
                offset=0,
                line_offset=0,
                max_indent=None,
            ).replace(**config),
        )
        return value

    def __len__(self):
        return self.len

    def __str__(self):  # pragma: no cover
        return self.to_string()


class Breakable(TextFormatter):
    def __init__(self, start, body, end):
        self.start = ensure_text(start)
        self.body = list(map(ensure_text, body))
        self.end = ensure_text(end)
        self.len = (
            len(self.start or "")
            + sum(map(len, self.body))
            + len(self.end or "")
        )

    def format(self, ctx):
        s = ""

        if self.start is not None:
            value, offset = ctx.format(self.start)
            s += value
        else:
            offset = ctx.offset

        if offset + len(self) - len(self.start or "") <= ctx.max_col:
            for x in self.body:
                value, offset = ctx.format(x, offset=offset)
                s += value

            if self.end is not None:
                value, offset = ctx.format(self.end, offset=offset)
                s += value
            return s, offset
        else:
            ind = ctx.next_line_offset()
            inner_ctx = ctx.replace(offset=ind, line_offset=ind)
            for x in self.body:
                value, offset = inner_ctx.format(x)
                s += "\n" + inner_ctx.indent() + value

            if self.end is not None:
                value, offset = ctx.format(self.end, offset=ctx.line_offset)
                s += "\n" + ctx.indent() + value

            return s, offset


class Sequence(TextFormatter):
    def __init__(self, *elements):
        self.elements = [Text(e) if isinstance(e, str) else e for e in elements]
        self.len = sum(map(len, self.elements))

    def format(self, ctx):
        s = ""
        offset = ctx.offset
        for e in self.elements:
            value, offset = ctx.format(e, offset=offset)
            s += value
        return s, offset


class Text(TextFormatter):
    def __init__(self, value):
        self.value = value
        self.len = len(self.value)

    def format(self, ctx):
        return self.value, ctx.offset + len(self)



"""Extensible HTML representation for Python objects."""

from .core import Config, Hrepr, HreprState, Interface, StdHrepr  # noqa: F401
from .h import HTML, H, Tag, css_hrepr  # noqa: F401
from .std import standard_html
from .term import standard_terminal

hrepr = Interface(StdHrepr, backend=standard_html)
trepr = Interface(StdHrepr, backend=standard_terminal)


def pprint(x, **config):  # pragma: no cover
    """Pretty-print an object.

    Arguments:
        indent: The number of spaces to use for indented blocks.
        max_col: The maximum width for the generated text.
        max_indent: The maximum indent (defaults to max_col - 15 if overflow
            is "allow").
        overflow: How to handle non-breakable text.
            * "allow" to let non-breakable text exceed max_col.
            * "break" to break the text on multiple indented lines.
            * "backslash" to break the text on multiple indented lines, with
                a leading backslash on each overflow line.
        config: Configuration settings (dependent on the pretty-printer for
            each object).
    """
    print(pstr(x, **config))


def pstr(x, indent=4, max_col=80, max_indent=None, overflow="allow", **config):
    """Convert x to a pretty-printed string.

    Arguments:
        indent: The number of spaces to use for indented blocks.
        max_col: The maximum width for the generated text.
        max_indent: The maximum indent (defaults to max_col - 15 if overflow
            is "allow").
        overflow: How to handle non-breakable text.
            * "allow" to let non-breakable text exceed max_col.
            * "break" to break the text on multiple indented lines.
            * "backslash" to break the text on multiple indented lines, with
                a leading backslash on each overflow line.
        config: Configuration settings (dependent on the pretty-printer for
            each object).
    """
    return trepr(x, **config).to_string(
        tabsize=indent,
        max_col=max_col,
        max_indent=max_indent,
        overflow=overflow,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hrepr']

package_data = \
{'': ['*'], 'hrepr': ['style/*']}

install_requires = \
['ovld>=0.2.7,<0.3.0']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.8,<0.9']}

setup_kwargs = {
    'name': 'hrepr',
    'version': '0.3.6',
    'description': 'Extensible HTML representation for Python objects.',
    'long_description': '# hrepr\n\n`hrepr` outputs HTML/pretty representations for Python objects.\n\n✅ Nice, colourful representations of lists, dicts, dataclasses, booleans...<br/>\n✅ Ridiculously extensible and configurable<br/>\n✅ Handles recursive data structures<br/>\n✅ Compatible with Jupyter notebooks<br/>\n✅ Also pretty-prints to the terminal<br/>\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr1.png" width="400px"><img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr2.png" width="400px">\n\nI suggest studying the two example files to learn `hrepr`:\n\n* `python examples/exhibit.py > exhibit.html` (and then view the HTML file)\n* `python examples/pretty.py`\n\nAlso see the Jupyter notebook at `examples/Basics.ipynb`, but keep in mind that GitHub can\'t display it properly because of the injected styles/scripts.\n\n\n## Install\n\n```python\npip install hrepr\n```\n\n\n## Usage\n\n```python\nfrom hrepr import hrepr\nobj = {\'potatoes\': [1, 2, 3], \'bananas\': {\'cantaloups\': 8}}\n\n# Print the HTML representation of obj\nprint(hrepr(obj))\n\n# Wrap the representation in <html><body> tags and embed the default\n# css style files in a standalone page, which is saved to obj.html\nhrepr.page(obj, file="obj.html")\n```\n\nIn a Jupyter Notebook, you can return `hrepr(obj)` from any cell and it will show its representation for you. You can also write `display_html(hrepr(obj))`.\n\n\n## Custom representations\n\nA custom representation for an object can be defined using the following three methods (it is not necessary to define all of them, only those that are relevant to your case):\n\n* `__hrepr__(self, H, hrepr)` returns the normal HTML representation.\n    * Use `H.span["some-class"](some-content, some_attr=some_value)` to generate HTML.\n    * Use `hrepr(self.x)` to generate the representation for some subfield `x`.\n    * `hrepr.config` contains any keyword arguments given in the top level call to `hrepr`. For instance, if you call `hrepr(obj, blah=3)`, then `hrepr.config.blah == 3` in all calls to `__hrepr__` down the line (the default value for all keys is `None`).\n* `__hrepr_short__(self, H, hrepr)` returns a *short* representation, ideally of a constant size.\n    * The output of this method is used when we hit max depth, or for repeated references.\n    * Only include bare minimum information. Short means short.\n* `__hrepr_resources__(cls, H)` is a **classmethod** that returns resources common to all instances of the class (typically a stylesheet or a script).\n    * When generating a page, the resources will go in `<head>`.\n    * You can return a list of resources.\n\nNo dependency on `hrepr` is necessary.\n\nFor example:\n\n```python\nclass Person:\n    def __init__(self, name, age, job):\n        self.name = name\n        self.age = age\n        self.job = job\n\n    @classmethod\n    def __hrepr_resources__(cls, H):\n        # Note: you might need to add "!important" next to some rules if\n        # they conflict with defaults from hrepr\'s own CSS.\n        return H.style("""\n            .person {\n                background: magenta !important;\n                border-color: magenta !important;\n            }\n            .person-short { font-weight: bold; color: green; }\n        """)\n\n    def __hrepr__(self, H, hrepr):\n        # H.instance is a special kind of tag to format data like an instance.\n        # Notice how we call the hrepr parameter on self.age and self.job to\n        # format them.\n        return H.instance["person"](\n            H.pair("age", hrepr(self.age), delimiter=" ↦ "),\n            H.pair("job", hrepr(self.job), delimiter=" ↦ "),\n            # The "type" represents the header for the "instance"\n            type=self.name,\n            # "vertical=True" means we\'ll display the pairs as a table with\n            # the delimiters aligned, instead of sticking them horizontally\n            # next to each other\n            vertical=True,\n        )\n\n    def __hrepr_short__(self, H, hrepr):\n        # H.atom is really mostly like H.span, but the textual representation\n        # of H.atom(x) through pprint is "x" whereas H.span(x) would be \n        # "<span>x</span>".\n        return H.atom["person-short"](self.name)\n```\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr3.png" width="600px">\n\nNote how this also gave us a textual representation *for free*, using `hrepr.pprint`. This feature requires using special tags like `H.instance`, `H.bracketed`, `H.pair` or `H.atom` -- other ones will pretty-print like HTML -- but it\'s neat regardless!\n\n\n## References\n\n`hrepr` (and `hrepr.pprint` for that matter) can handle circular references. Furthermore, if an object is found at several places in a structure, only the first occurrence will be printed in full, and any other will be a numeric reference mapped to the short representation for the object. It looks like this:\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr4.png" width="600px">\n\nThe `shortrefs` and `norefs` configuration keys control the representation of references:\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr5.png" width="600px">\n\n`norefs` is ignored when there are circular references.\n\n\n## HTML generation\n\nGenerate HTML using the `H` parameter to `__hrepr__`, or import it and use it directly:\n\n```python\nfrom hrepr import H\nhtml = H.span["bear"](\n    "Only ", H.b("YOU"), " can prevent forest fires!",\n    style="color: brown;"\n)\nprint(html)\n# <span class="bear" style="color: brown;">Only <b>YOU</b> can prevent forest fires!</span>\n```\n\n`H` can be built incrementally: if you have an element, you can call it to add children, index it to add classes, and so on. For instance:\n\n```python\nfrom hrepr import H\nhtml = H.span()\nhtml = html("Only ")\nhtml = html(style="color: brown;")["bear"]\nhtml = html(H.b("YOU"), " can prevent forest fires!")\nprint(html)\n# <span class="bear" style="color: brown;">Only <b>YOU</b> can prevent forest fires!</span>\n```\n\nThis can be handy if you want to tweak generated HTML a little. For example, `hrepr(obj)["fox"]` will tack on the class `fox` to the representation of the object.\n\n\n### Special tags\n\nStandard tags like `H.span`, `H.div`, `H.strong`, etc. are handled according to standards. But there are some special tags which are postprocessed by the hrepr "backend". The HTML backend will reduce them to standard tags, whereas the pprint backend will display them like Python data structures.\n\n* `H.instance(*children, type=<str>, delimiter=<str>, short=<bool>, horizontal=<bool>, vertical=<bool>)`\n    * Represents some kind of object\n    * `type`: the name of the object (or class name)\n        * It is not *necessarily* a string, it can also be a tag.\n    * `delimiter`: the delimiter between elements, defaults to a comma. The HTML formatter ignores this.\n    * `short/horizontal/vertical`: the layout/style\n* `H.bracketed(*children, type=<str>, open=<str>, close=<str>, delimiter=<str>, short=<bool>, horizontal=<bool>, vertical=<bool>)`\n    * `type`: the name of the object (or class name), which is NOT displayed. Instead, the class `hreprt-<name>` is given to the element.\n    * `open/close`: the opening and closing brackets.\n    * `delimiter`: the delimiter between elements, defaults to a comma. The HTML formatter ignores this.\n    * `short/horizontal/vertical`: the layout/style. Lists use horizontal, dicts use vertical, and the short representations use short.\n* `H.pair(x, y, delimiter=<str>)`: a key -> value mapping. They are handled specially inside of `bracketed` and `instance` so that the delimiters are aligned.\n* `H.atom(element, type=<str>)`: essentially equivalent to `H.span["hreprt-<type>"](element)`, or `repr(element)` for pprint.\n\n\n### Including JavaScript libraries\n\nTo make it a bit easier to include and use JavaScript libraries, there is a new tag called `H.javascript` that uses RequireJS under the hood:\n\n* `H.javascript(export=<name>, src=<path>)` declares an import of the script at the given path (you can use a CDN link) and exports it under the given name.\n* `H.javascript(<code>, require=<name/list>, export=<name>)` will wait for the required modules (the names in the require list are the ones given in the export field of other javascript tags) and will provide them to your script.\n\nOptionally you can name a variable as the export, so that it can be required by other scripts.\n\n\n### Constructed elements\n\nEach element gains a new attribute, `constructor`, which must name a function exported using a `javascript` tag. If provided, the function is called with the element as the first argument, and the `options` attribute as the second element.\n\nTo illustrate, here\'s an example of requiring and using Plotly:\n\n```python\nclass Plot:\n    def __init__(self, data):\n        self.data = data\n\n    @classmethod\n    def __hrepr_resources__(cls, H):\n        return [\n            H.javascript(\n                export="plotly",\n                src="https://cdn.plot.ly/plotly-latest.min.js",\n            ),\n            H.javascript(\n                """\n                function PlotlyPlot(element, data) {\n                    plotly.newPlot(element, data);\n                }\n                """,\n                require="plotly",\n                export="PlotlyPlot",\n            ),\n        ]\n\n    def __hrepr__(self, H, hrepr):\n        return H.div(\n            constructor="PlotlyPlot",\n            options=[{"x": list(range(len(self.data))), "y": list(self.data)}],\n        )\n```\n\n* `__hrepr_resources__` declares two resources:\n  * The plotly library, loaded from a CDN, and exported as `plotly`.\n  * A small snippet of code that requires `plotly` and declares a `PlotlyPlot` function that takes an element and some data.\n* `__hrepr__` returns a div that has `PlotlyPlot` as the constructor. It will be called as `new PlotlyPlot(element, options)` (this is so it may equally be defined as an ES6 class if you wish). The options serialized as JSON and dumped into the call -- they are the "data" argument in make_plot.\n\nWe can then test it:\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr6.png" width="600px">\n\nAlternatively, you can define custom elements using the [custom element API](https://developer.mozilla.org/en-US/docs/Web/Web_Components/Using_custom_elements). You can put the definition in a `<javascript>` tag that requires the library you want to use, and then simply use `H.my_element(...)` to instantiate it. The underscore in the Python attribute will become a dash, so hrepr will generate `<my-element>...</my-element>`.\n\n## Customize hrepr\n\n### Mixins\n\nIf you want to *really* customize hrepr, you can use mixins. They look like a bit of black magic, but they\'re simple enough:\n\n```python\n# ovld is one of the dependencies of hrepr\nfrom ovld import ovld, has_attribute\nfrom hrepr import hrepr, Hrepr\n\nclass MyMixin(Hrepr):\n    # Change the representation of integers\n\n    def hrepr_resources(self, cls: int):\n        # Note: in hrepr_resources, cls is the int type, not an integer\n        return self.H.style(".my-integer { color: fuchsia; }")\n\n    def hrepr(self, n: int):\n        return self.H.span["my-integer"]("The number ", str(n))\n\n    # Specially handle any object with a "quack" method\n\n    def hrepr(self, duck: has_attribute("quack")):\n        return self.H.span("🦆")\n```\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr7.png" width="600px">\n\nThe annotation for a rule can either be a type, `ovld.has_attribute`, or pretty much any function wrapped with the `ovld.meta` decorator, as long as the function operates on classes. See the documentation for [ovld](https://github.com/breuleux/ovld#other-features) for more information.\n\nAnd yes, you can define `hrepr` multiple times inside the class, as long as they have distinct annotations and you inherit from `Hrepr`. You can also define `hrepr_short` or `hrepr_resources` the same way.\n\n### Postprocessors\n\n`hrepr` can be given a postprocessor that is called on the representation of any object. You can use this to do things like highlighting specific objects:\n\n```python\nfrom hrepr import H\n\nstyle = H.style(".highlight { border: 3px solid red !important; }")\n\ndef highlight(x):\n    def postprocess(element, obj, hrepr):\n        if obj == x:\n            # Adds the "highlight" class and attaches a style\n            return element["highlight"].fill(resources=style)\n        else:\n            return element\n\n    return postprocess\n\nhrepr([1, 2, [3, 4, 2]], postprocess=highlight(2))\n```\n\n<img src="https://raw.githubusercontent.com/breuleux/hrepr/master/images/hrepr8.png" width="600px">\n\n\n### hrepr variants\n\nTo put this all together, you can create a *variant* of `hrepr`:\n\n```python\nhrepr2 = hrepr.variant(mixins=MyMixin, postprocess=highlight(2))\nhrepr2([1, 2, 3])  # Will use the mixins and postprocessor\n```\n\n\n### Configure the hrepr function itself\n\nAlternatively, you can configure the main `hrepr`:\n\n```python\nhrepr.configure(mixins=MyMixin, postprocess=highlight(2))\n```\n\nBut keep in mind that unlike the variant, the above will modify `hrepr` for everything else as well.\n',
    'author': 'Olivier Breuleux',
    'author_email': 'breuleux@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/breuleux/hrepr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from ovld import ovld

from .h import H, Tag
from .textgen import Breakable, Sequence, Text, join


@ovld
def standard_terminal(self, node: Tag):
    """Generate a terminal representation."""
    return node.text_parts()


@ovld
def standard_terminal(self, node: type(H.atom)):
    typ = node.get_attribute("type", None)
    if typ == "str":
        (child,) = node.children
        return Text(repr(child))
    else:
        return Text("".join(map(str, node.children)))


@ovld
def standard_terminal(self, node: type(H.defn)):
    key, name = node.children
    return Text(f"{key} {name}")


@ovld
def standard_terminal(self, node: type(H.pair)):
    delim = node.get_attribute("delimiter", ": ")
    k, v = node.children
    return Sequence(self(k), delim, self(v))


@ovld
def standard_terminal(self, node: type(H.bracketed)):
    start = node.get_attribute("start", "(")
    end = node.get_attribute("end", ")")
    delim = node.get_attribute("delimiter", ", ")
    return Breakable(
        start=start, body=join(map(self, node.children), delim), end=end,
    )


@ovld
def standard_terminal(self, node: type(H.instance)):
    typ = node.get_attribute("type", "<object>")
    delim = node.get_attribute("delimiter", ", ")
    start = node.get_attribute("start", "(")
    end = node.get_attribute("end", ")")
    return Breakable(
        start=Sequence(self(typ), start),
        body=join(map(self, node.children), delim),
        end=end,
    )


@ovld
def standard_terminal(self, node: type(H.ref)):
    num = node.get_attribute("num", -1)
    ref = Sequence("#", str(num))
    if node.children:
        return Sequence(ref, "=", *map(self, node.children))
    else:
        return ref


@ovld
def standard_terminal(self, node: object):
    return str(node)
